// ABgestureShell - This is a DIVERSE DSO which loads files
// containing pairs of objects and corresponding commands, specified 
// by an environment variable, DPF_ABgestureShell_FILES. 
// Originally, the first of each pair of objects is displayed. 
// When the wand is pointed at an object,
// its partner object will be displayed instead. The wand MUST be in a
// faced-down position. In order to choose the next object, the wand
// must be returned to an upright position and then flipped down again.
//
// A second environment variable, ABgestureShell_A_ON, can be used to keep
// the first object in each set permanently turned on. Set this variable
// to 'on.' If this variable is not present, if or it is not set to 'on,'
// the objects will flip from one to the other only.
//
// Authors:
//     Adele Peskin
//      Scientific Applications & Visualization Group
//           SAVG web page: http://math.nist.gov/mcsd/savg/
//      Mathematical & Computational Science Division
//           MCSD web page: http://math.nist.gov/mcsd/
//      National Institute of Standards and Technology
//           NIST web page: http://www.nist.gov/
//      
//     John Kelso
//       Virginia Tech, http://www.diverse.vt.edu/, http://www.cave.vt.edu/
//   
//    Steve Satterfield (SAVG) and John Kelso provided 
//    the idea for the program
//
//
// For information about this DSO, 
// contact Adele Peskin, peskin@boulder.nist.gov
//
//
// This code is a modification to the example generic dtkAugment DSO provided
// with the DIVERSE distribution. The following disclaimer applies to the NIST
// modifications.
// 
// "This software was developed at the National Institute of Standards
// and Technology by employees of the Federal Government in the course
// of their official duties. Pursuant to title 17 Section 105 of the
// United States Code this software is not subject to copyright
// protection and is in the public domain.  xxxxxxx is an experimental 
// system.  NIST assumes no responsibility whatsoever for its use by 
// other parties, and makes no guarantees, expressed or implied, about 
// its quality, reliability, or any other characteristic.
// 
// We would appreciate acknowledgement if the software is used."

#include <stdio.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>
#include <fstream.h>
#include "selectVisible.C"
#include "testSuffix.C"
#include "parse.C"

class ABgestureShell : public dtkAugment {
public:
  ABgestureShell(dpf *app) ;
  ~ABgestureShell(void) ;
  int postConfig(void);
  int postFrame(void);
  
private:
  dpf* app;
  dpfDisplay *display;
  dpfToggle** mygroupA;
  dpfToggle** mygroupB;
  //  Shared memory segment for the spin rate
  dtkInLocator *wand ;
  //  Shared memory segment used to request the wand pointer
  dtkSharedMem* shm2;
  //  Shared memory segment used to request a system call
  dtkSharedMem* shm4;
  //  Shared memory segment to put the call
  dtkSharedMem* shm_call;
  int cnt;
  unsigned char selected;
  unsigned char newCallSet;
  unsigned char* toggleChars;
  char** commandListA;
  char** commandListB;
  int* Aon;
  pfNode** objectPtrsA;
  pfNode** objectPtrsB;
  pfGroup* rootGroup;
  pfCoord wandCoord;
  pfVec3 check;
  void resetAorB();
  float wandData[6];
  int leaveAon;
  bool oldWandup;
  bool wandup;
  bool readyToSelect;
  bool makeSound;
  int flipcnt;
  int last_one_hit;
  char* file_a;
  char* file_b; 
  char* commndA; 
  char* commndB;
  void testForNull();
} ;

// create object and set description.
// be sure to validate it!
ABgestureShell::ABgestureShell(dpf *app_in) :
  dtkAugment("ABgestureShell")
{
  setDescription("loads files specified by DPF_ABSWITCH_FILES"
		 " environment variable and display one of two"
		 " sets of data files, set A and set B. A second"
		 " mode can be used to keep set A on all the time,"
		 " by setting a second environment variable,"
		 " ABSWITCH_A_ON to 'on'"
		 " This version was compiled on 5/20/03.") ;
  app = app_in;
  validate();
  cnt = 0 ;
  objectPtrsA = NULL;
  objectPtrsB = NULL;
  commandListA = NULL;
  commandListB = NULL;
  Aon = NULL;
  leaveAon = 0;
  wand = (dtkInLocator *) app->get("wand", DTKINLOCATOR_TYPE);
  if(!wand || wand->isInvalid())
    wand = NULL ;

  shm2 = new dtkSharedMem(sizeof(unsigned char), "pointerOn");
  if(!shm2 || shm2->isInvalid()) {
      shm2 = NULL;
  }

  shm4 = new dtkSharedMem(sizeof(unsigned char), "callFlag3");
  if(!shm4 || shm4->isInvalid()) {
      shm4 = NULL;
  }

  newCallSet = 1;
  if (shm4)
    shm4->write(&newCallSet);

  shm_call = new dtkSharedMem(sizeof(char)*100, "sysCall3");
  if(!shm_call || shm_call->isInvalid())
  shm_call = NULL ;

  oldWandup = true;
  wandup = true;
  readyToSelect = false;
  flipcnt = 1;
  last_one_hit = -1;
  makeSound = false;
} ;

// destructor
ABgestureShell::~ABgestureShell(void) {
  newCallSet = 3;
  if (shm4)
	shm4->write(&newCallSet);
  dtkSharedMem_destroy("pointerOn");
  dtkSharedMem_destroy("callFlag3");
  dtkSharedMem_destroy("sysCall3");
}


int ABgestureShell::postConfig(void) {
  
  // init file loaders
  char *env = getenv("DPF_ABSWITCH_FILES") ;
  char *addnoise = getenv("ADD_NOISE") ;
  char *original_list ;
  char *line = new char[2047];
  file_a = new char[100];
  file_b = new char[100];
  commndA = new char[100];
  commndB = new char[100];
  int pargc ;
  char **pargv ;
  char *pOrm = new char[1]; ;

  if (env) {
    original_list = strdup(env) ;
    // strtok() is broken in GNU libc, so we roll our own here.
    char *end = original_list ;
    for(;*end;) {
      for(; *end != ':' && *end; end++);
      if(*end) { *end = '\0'; end++; }
      // add to the list of files to read
      if (strlen(original_list)>0) {
	// read each file and fill the filea and fileb arrays with names
	FILE *fp = fopen(original_list, "r");        
        if (!fp) {
	  dtkMsg.add(DTKMSG_ERROR,
		     "ABgestureShell::preConfig: error opening input file %s\n",
		     original_list) ;
	  return DTK_CALLBACK_ERROR;
        }
	
	//printf("next file off of the list \"%s\"\n",original_list) ;
	while ((fgets(line,250,fp)!=NULL)) {
	  //cout << "got the next line: " << line << endl;
	  // parse the line into 2 objects and a command if there
	  dpf_charsToArgcArgv(line, &pargc, &pargv) ;

	  // this line is good if there are 3 parameters
	  //printf("after parse, pargc=%d\n",pargc) ;
	  if (pargc == 5) {
	    leaveAon = 0;
	    pOrm = pargv[0];
	    if (pOrm[0] == '+') leaveAon = 1;
	    if (pOrm[0] == '-') leaveAon = -1;
	    file_a = pargv[1];
	    file_b = pargv[2];
	    commndA = pargv[3];
	    commndB = pargv[4];
	    //cout << "after parse: leaveAon: " << leaveAon << " filea: " << file_a << "  " << file_b << "  " << commndA << endl;
	    objectPtrsA = (pfNode **) realloc(objectPtrsA, sizeof(pfNode*)*(cnt+1));
	    objectPtrsB = (pfNode **) realloc(objectPtrsB, sizeof(pfNode*)*(cnt+1));
	    commandListA = (char **) realloc(commandListA, sizeof(char*)*(cnt+1));
	    commandListB = (char **) realloc(commandListB, sizeof(char*)*(cnt+1));
	    // test suffixes and store files
	    testSuffix(file_a);
	    objectPtrsA[cnt] = pfdLoadFile(file_a);
	    testForNull();
	    if (file_b == NULL) {
	      objectPtrsB[cnt] = NULL;
	    }
	    else {
	      testSuffix(file_b);
	      objectPtrsB[cnt] = pfdLoadFile(file_b);
	    }
	    commandListA[cnt] = commndA;
	    commandListB[cnt] = commndB;
	    commndA = new char[100];
	    commndB = new char[100];
	    // set the leave a on array for this set of objects
	    Aon = (int *) realloc(Aon, sizeof(int)*(cnt+1));
	    Aon[cnt] = leaveAon;
	    cnt++ ;
	  }
	  else {
	    file_a = line;
	    cout << "Incorrect line: " << file_a << endl;
	    dtkMsg.add(DTKMSG_ERROR,
		       "ABswitchShell:postConfig: input file error :"
		       "This line does not have the correct 5 parameters.\n");
	  }
	}
	original_list = end;
      }
    }
  }

  display = app->display();
  toggleChars = new unsigned char[cnt];
  mygroupA = new dpfToggle*[cnt] ;
  mygroupB = new dpfToggle*[cnt] ;

  rootGroup = new pfGroup;
  display->world()->addChild(rootGroup);
  
  char *old_path = (char *) pfGetFilePath();
  char *path = NULL;
  if(old_path)
    {
      // copy old_path from static memory to the heap.
      old_path = dtk_strdup(old_path);
      path = (char *) dtk_malloc(strlen(old_path) +
				 strlen("%s:.:"DPF_DATA_DIR));
      sprintf(path,"%s:.:"DPF_DATA_DIR, old_path);
      pfFilePath(path);
    }
  else // old_path == NULL
    pfFilePath(".:"DPF_DATA_DIR);

  for (int i = 0; i< cnt; i++) {
    // define a toggle node for each object that is loaded
    toggleChars[i] = 2;
    mygroupA[i] = new dpfToggle(&toggleChars[i], 1) ;
    mygroupB[i] = new dpfToggle(&toggleChars[i], 0) ;

    // add objects A and B to the toggleNode
    mygroupA[i]->addChild(objectPtrsA[i]) ;
    mygroupB[i]->addChild(objectPtrsB[i]) ;
    rootGroup->addChild(mygroupA[i]);
    rootGroup->addChild(mygroupB[i]);
  }

  return DTK_REMOVE_CALLBACK ; 
  }

void ABgestureShell::testForNull()
{
  // test file_b
  if (file_b[0]=='N'&&file_b[1]=='U'&&file_b[2]=='L'&&file_b[3]=='L') {
    file_b = NULL;
  }
  // test the commands
  if (commndA[0]=='N'&&commndA[1]=='U'&&commndA[2]=='L'&&commndA[3]=='L') { 
    commndA = NULL;
  }
  if (commndB[0]=='N'&&commndB[1]=='U'&&commndB[2]=='L'&&commndB[3]=='L') { 
    commndB = NULL;
  }
}

// called every display loop, as a pfFrame callback, after pfDraw()
// this return value causes it to only be called once, and the object removed
int ABgestureShell::postFrame(void) {
  char* cmd;
  // get the toggle from the direction of the wand
  if (wand)
    wand->read(wandData);
  // create a matrix from the HPR position
  pfMatrix wandMat;
  wandMat.makeEuler(wandData[3],wandData[4],wandData[5]);

  // define the wand data in world coordinates
  wandCoord.xyz.set(wandData[0], wandData[1], wandData[2]);
  wandCoord.hpr.set(wandData[3], wandData[4], wandData[5]);
  display->dpfToWorld(&wandCoord);
    
  // create a matrix from the HPR position in world coordinates
  pfMatrix wandWorld;
  wandWorld.makeEuler(wandCoord.hpr[0],wandCoord.hpr[1],wandCoord.hpr[2]);

  // create a vector in the +y direction and move according to the wand
  pfVec3 upVector(0.f, 1.f, 0.f);
  check.xformVec(upVector, wandWorld);

  // first see if the wand is upside down - use the upside
  // down wand to point to objects

  // create a vector in the +z direction
  pfVec3 zVector(0.f, 0.f, 1.f);

  // move it according to the world wand position
  zVector.xformVec(zVector, wandWorld);

  // each time an object is selected the wand should flip down
  // again. Keep track of whether the wand is up or down
  oldWandup = wandup;
  if (zVector[PF_Z] > 0.f) wandup = true;
  else wandup = false;

  // see if the wand just flipped
  if ((oldWandup && !wandup) || (!oldWandup && wandup)) {
    ++flipcnt;
    if (flipcnt > 2) flipcnt -= 2;
    if (flipcnt == 2 ) readyToSelect = true;
    else readyToSelect = false;
  }

  // go on to the object selection if readyToSelect is true
  if (!readyToSelect) return DTK_CONTINUE ;
  // make the wandPointer appear
  if (shm2) {
    selected = 1;
    shm2->write(&selected);
  }

  // do the intersection and find the visible object that is closest
  // to the wand
  // the next routine reset the last_one_hit object number if found
  int last = selectVisible(cnt, toggleChars, check, 
			   wandCoord, objectPtrsA, objectPtrsB);
  // get new value of newCallSet
  if (shm4)
    shm4->read(&newCallSet);
  if (last > -1 && newCallSet == 1) {
    last_one_hit = last;
    // reset the toggle
    resetAorB();
    // change the setting of readyToSelect
    readyToSelect = false;
    flipcnt = 0;

    // make the wandPointer disappear
    if (shm2) {
      selected = 2;
      shm2->write(&selected);
    }

      // see if there is a command for this object
      cmd = commandListA[last_one_hit];
      if ( toggleChars[last_one_hit] == 2) {
	cmd = commandListB[last_one_hit];
      }
      // cout << "got the selected command:newcallset back to 2 " << cmd << endl;

    // as long as the call is not NULL, 
    // read it out so dtk-runCommand can run it
    if (cmd != NULL) {
      // set the flag to tell the other program to read a call
      newCallSet = 2;
      if (shm4)
	shm4->write(&newCallSet);
      if (shm_call) {
	shm_call->write(cmd);
      }
    }
  }

  return DTK_CONTINUE ;
}


void ABgestureShell::resetAorB() {
  // the pair to switch is number: last_one_hit
  leaveAon = Aon[last_one_hit];

  // either: leaveAon = 1: object 1 always on
  if (leaveAon == 1) {
    if (toggleChars[last_one_hit] == 3) 
      toggleChars[last_one_hit] = 2;
    else
      toggleChars[last_one_hit] = 3;
  }

  // or: leaveAon = -1: switch whatever the case
  if (leaveAon == -1) {
    if (toggleChars[last_one_hit] == 2) {
      toggleChars[last_one_hit] = 1;
    }
    else {
      toggleChars[last_one_hit] = 2;
    }
  }

}

/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new ABgestureShell(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
