/* this is an intersection routine that selects the objects
   of toggleNodes that are closest to a wand pointer. Only
   visible objects can be selected.
   Input:
   cnt - number of toggleNodes holding objects
   toggleChars - bits that say which object is visible
   check - direction vector of the wand
   wandCoord - coordinates of the wand from a world point of view
   objectPtrsA - pointers to the first objects in the toggleNodes
   objectPtrsB - pointers to the second objects in the toggleNodes
   -------------------------------------------------------------*/   

int selectVisible(int cnt, unsigned char* toggleChars, pfVec3 check,
		   pfCoord wandCoord, 
		  pfNode** objectPtrsA, pfNode** objectPtrsB) {
  int last_one_hit = -1;

  // for each visible object, see if the check vector is within
  // its bounding sphere, and keep track of the one that is
  // the closest to the wand

  // loop through each pair and find the visible one
  pfNode* node;
  pfSphere *bsph = new pfSphere();
  float objRadius;
  pfVec3 centerPt;
  pfVec3 intPt;
  float distToWand = 1000000.0;
  for ( int j = 0; j < cnt; j++ ) {
    node = objectPtrsA[j];
    if (toggleChars[j] == 1 && objectPtrsB[j] != NULL) { 
      node = objectPtrsB[j];
    }
    if (node == NULL) continue;
    
    // find bounding sphere
    node->getBound(bsph);
    objRadius = bsph->radius;
    centerPt = bsph->center;

    // Is the distance from centerPt to the wand is less than the radius?

    // This code is from the objectMover: OMS.C
    // the line from the wand is defined by a point(wandCoord) 
    // and a direction (check) r = r0 + d(direction);
    // so, need to find the distance d up the line that is perpendicular
    // to the center point, that is at the point P=(a,b,c)

    // the dot product of the P-centr with P-wand = 0 
    // if the vectors are perpendicular
    // did some algebra and came up with an expression for d:
    float d = check[0]*(centerPt[0]-wandCoord.xyz[0]) +
              check[1]*(centerPt[1]-wandCoord.xyz[1]) +
              check[2]*(centerPt[2]-wandCoord.xyz[2]);
    d = d/(check[0]*check[0] + check[1]*check[1] + check[2]*check[2]);

    // this gives the point of intersection with the wand, int: now find
    // the distance from the centerpoint to int
    intPt[0] = wandCoord.xyz[0] + d*check[0];
    intPt[1] = wandCoord.xyz[1] + d*check[1];
    intPt[2] = wandCoord.xyz[2] + d*check[2];
    float dist = sqrt((centerPt[0]-intPt[0])*(centerPt[0]-intPt[0]) +
		      (centerPt[1]-intPt[1])*(centerPt[1]-intPt[1]) +
		      (centerPt[2]-intPt[2])*(centerPt[2]-intPt[2]));

    // see if the sphere hits the wand
    if (dist < objRadius && d > 0) {
      // this object hits!
      // save distance = d
      if (dist < distToWand) {
	distToWand = dist;
	last_one_hit = j;
      }
    }
  }

  return last_one_hit;
}


// this is the corresponding routine for object not toggled
int selectClosest(int cnt, pfVec3 check, pfCoord wandCoord, 
		  pfNode** objectPtrsA) {
  int last_one_hit = -1;

  // for each visible object, see if the check vector is within
  // its bounding sphere, and keep track of the one that is
  // the closest to the wand

  // loop through each pair and find the visible one
  pfNode* node;
  pfSphere *bsph = new pfSphere();
  float objRadius;
  pfVec3 centerPt;
  pfVec3 intPt;
  float distToWand = 1000000.0;
  for ( int j = 0; j < cnt; j++ ) {
    node = objectPtrsA[j];
    if (node == NULL) continue;

    // find bounding sphere
    node->getBound(bsph);
    objRadius = bsph->radius;
    centerPt = bsph->center;

    // Is the distance from centerPt to the wand is less than the radius?

    // This code is from the objectMover: OMS.C
    // the line from the wand is defined by a point(wandCoord) 
    // and a direction (check) r = r0 + d(direction);
    // so, need to find the distance d up the line that is perpendicular
    // to the center point, that is at the point P=(a,b,c)

    // the dot product of the P-centr with P-wand = 0 
    // if the vectors are perpendicular
    // did some algebra and came up with an expression for d:
    float d = check[0]*(centerPt[0]-wandCoord.xyz[0]) +
              check[1]*(centerPt[1]-wandCoord.xyz[1]) +
              check[2]*(centerPt[2]-wandCoord.xyz[2]);
    d = d/(check[0]*check[0] + check[1]*check[1] + check[2]*check[2]);

    // this gives the point of intersection with the wand, int: now find
    // the distance from the centerpoint to int
    intPt[0] = wandCoord.xyz[0] + d*check[0];
    intPt[1] = wandCoord.xyz[1] + d*check[1];
    intPt[2] = wandCoord.xyz[2] + d*check[2];
    float dist = sqrt((centerPt[0]-intPt[0])*(centerPt[0]-intPt[0]) +
		      (centerPt[1]-intPt[1])*(centerPt[1]-intPt[1]) +
		      (centerPt[2]-intPt[2])*(centerPt[2]-intPt[2]));

    // see if the sphere hits the wand
    if (dist < objRadius && d > 0) {
      // this object hits!
      // save distance = d
      if (dist < distToWand) {
	distToWand = dist;
	last_one_hit = j;
      }
    }
  }

  return last_one_hit;
}


int selectToggles(int cnt, unsigned char* toggleChars, pfVec3 check,
		   pfCoord wandCoord, 
		  pfNode** objectPtrsA, pfNode** objectPtrsB) {
  int last_one_hit = -1;

  // for each visible object, see if the check vector is within
  // its bounding sphere, and keep track of the one that is
  // the closest to the wand

  // loop through each pair and find the visible one
  pfNode* node;
  pfNode* node2;
  pfSphere *bsph = new pfSphere();
  float objRadius;
  pfVec3 centerPt;
  pfVec3 intPt;
  float distToWand = 1000000.0;
  int times;
  for ( int j = 0; j < cnt; j++ ) {
    // there should always be at least one object or the other:
    // if both are visible, test both
    // if there is only one: it should always be visible
    times = 1;
    node2 = NULL;
    if (objectPtrsB[j] == NULL)
      node = objectPtrsA[j];
    else {
      if (toggleChars[j] == 1)
	node = objectPtrsB[j];
      if (toggleChars[j] == 2)
	node = objectPtrsA[j];
      if (toggleChars[j] == 3) {
	node = objectPtrsA[j];
	node2 = objectPtrsB[j];
	times = 2;
      }
    }
    if (node == NULL) continue;
    for ( int k = 0; k < times; ++k ) {
      if (k==1) node = node2;

      // find bounding sphere
      node->getBound(bsph);
      objRadius = bsph->radius;
      centerPt = bsph->center;

      // Is the distance from centerPt to the wand is less than the radius?
      
      // This code is from the objectMover: OMS.C
      // the line from the wand is defined by a point(wandCoord) 
      // and a direction (check) r = r0 + d(direction);
      // so, need to find the distance d up the line that is perpendicular
      // to the center point, that is at the point P=(a,b,c)
      
      // the dot product of the P-centr with P-wand = 0 
      // if the vectors are perpendicular
      // did some algebra and came up with an expression for d:
      float d = check[0]*(centerPt[0]-wandCoord.xyz[0]) +
	check[1]*(centerPt[1]-wandCoord.xyz[1]) +
	check[2]*(centerPt[2]-wandCoord.xyz[2]);
      d = d/(check[0]*check[0] + check[1]*check[1] + check[2]*check[2]);
      
      // this gives the point of intersection with the wand, int: now find
      // the distance from the centerpoint to int
      intPt[0] = wandCoord.xyz[0] + d*check[0];
      intPt[1] = wandCoord.xyz[1] + d*check[1];
      intPt[2] = wandCoord.xyz[2] + d*check[2];
      float dist = sqrt((centerPt[0]-intPt[0])*(centerPt[0]-intPt[0]) +
			(centerPt[1]-intPt[1])*(centerPt[1]-intPt[1]) +
			(centerPt[2]-intPt[2])*(centerPt[2]-intPt[2]));
      
      // see if the sphere hits the wand
      if (dist < objRadius && d > 0) {
	// this object hits!
	// save distance = d
	if (dist < distToWand) {
	  distToWand = dist;
	  last_one_hit = j;
	}
      }
    }
  }

  return last_one_hit;
}
